/* Process data from the DCT2000 and fill data structures accordingly.
 * 
 * Based highly on DCTTest by Don Starr
 * http://www.starrsoft.com/freeware/DCTTest.zip
 */ 

#include <stdio.h>
#include <stdlib.h>
#include <err.h>
#include <string.h>
#include <stdint.h>

#include "crc.h"
#include "packet.h"
#include "debug.h"

int seq_ack, seq_req;

_sProcessorInfo *ProcessorInit( void )
{
	_sProcessorInfo *retval = (_sProcessorInfo *)malloc( sizeof(_sProcessorInfo) );
	if ( retval != NULL )
	{
		retval->state = state_WaitPacketType;
		retval->packet.cpPayload = NULL;
		retval->payloadOffset = 0;
		retval->escReceived = 0;
	}
	
	return retval;
}

void ProcessorCleanup( _sProcessorInfo *info )
{
	if ( info != NULL )
	{
		// free any previously-allocated payload buffer
		if ( info->packet.cpPayload != NULL )
		{
			free( info->packet.cpPayload );
			info->packet.cpPayload = NULL;
		}

		free( info );
	}
}

int ProcessByte( _sProcessorInfo *info, unsigned char ch )
{
	// This code processes individual bytes, running a little state machine
	// to decode complete packets.
	// It expects "complete" packets - i.e. it's not currently set up to handle
	// start-of-packet (x10 x78) or end-of-packet (x10 x03) indications within
	// a partially-decoded packet.
	//
	// Returns 1 if packet is fully decoded, 0 otherwise.

	int retval = 0;

	if ( ch == 0x10 && !info->escReceived) {
		info->escReceived = 1;
		return 0;
	}

	switch ( info->state )
	{
	default:
		info->state = state_WaitPacketType;

	case state_WaitPacketType:
		if ( info->escReceived )
		{
			info->packet.cPacketType = ch;
			info->state = state_WaitLenHi;
		}
		break;

	case state_WaitLenHi:
		info->packet.iLen = (ch << 8);
		info->state = state_WaitLenLo;
		break;

	case state_WaitLenLo:
		info->packet.iLen |= ch;
		info->state = state_WaitSeq;
		break;

	case state_WaitSeq:
		info->packet.cSeq = ch;
		info->state = state_WaitID;
		break;

	case state_WaitID:
		info->packet.cID = ch;
		// free any previously-allocated payload buffer
		if ( info->packet.cpPayload != NULL )
		{
			free( info->packet.cpPayload );
			info->packet.cpPayload = NULL;
		}
		if ( info->packet.iLen - 2 )
		{
			info->packet.cpPayload = (unsigned char *)malloc( info->packet.iLen - 2 );
			info->payloadOffset = 0;
			info->state = state_WaitPayload;
		}
		else
		{
			info->state = state_WaitCRCHi;
		}
		break;

	case state_WaitPayload:
		info->packet.cpPayload[ info->payloadOffset ] = ch;
		if ( ++info->payloadOffset == info->packet.iLen-2 )
		{
			info->state = state_WaitCRCHi;
		}
		break;

	case state_WaitCRCHi:
		info->packet.iCRC = (ch << 8);
		info->state = state_WaitCRCLo;
		break;

	case state_WaitCRCLo:
		info->packet.iCRC |= ch;
		info->state = state_WaitTerminator;
		break;

	case state_WaitTerminator:
		if ( info->escReceived )
		{
			uint16_t crc = 0;

			crc = makecrc(crc, &info->packet.cPacketType, 1);
			crc = makecrc(crc, ((uint8_t *)&info->packet.iLen)+1, 1); /* MSB */
			crc = makecrc(crc, ((uint8_t *)&info->packet.iLen)+0, 1); /* LSB */
			crc = makecrc(crc, &info->packet.cSeq, 1);
			crc = makecrc(crc, &info->packet.cID, 1);
			crc = makecrc(crc, info->packet.cpPayload, info->packet.iLen-2);

			if ( crc != info->packet.iCRC ) 
				warnx("Bad CRC: expected 0x%04hx, got 0x%04hx",crc,info->packet.iCRC);

			retval=1;
		}

		info->state = state_WaitPacketType;
		break;
	}

	info->escReceived = 0;

	return retval;
}

/* Cleanup the payload memory */
void packet_cleanup(_sPacket *p)
{
	if(p!=NULL && p->cpPayload!=NULL) {
		free(p->cpPayload);
		p->cpPayload=NULL;
	}
}

void packet_alloc_payload(_sPacket *p)
{
	p->cpPayload = (uint8_t *)malloc(p->iLen-2);
	if(p->cpPayload==NULL)
		err(1,"malloc");
}

void packet_fill_crc(_sPacket *p)
{
	uint16_t crc = 0;

	crc = makecrc(crc, &p->cPacketType, 1);
	crc = makecrc(crc, ((uint8_t *)&p->iLen)+1, 1); /* MSB */
	crc = makecrc(crc, ((uint8_t *)&p->iLen)+0, 1); /* LSB */
	crc = makecrc(crc, &p->cSeq, 1);
	crc = makecrc(crc, &p->cID, 1);
	crc = makecrc(crc, p->cpPayload, p->iLen-2);

	p->iCRC = crc;
}

int response_parse_channel(_sPacket *p)
{
	printf("length is %d\n",p->iLen);
	return -1;
}

void packet_build_channelquery(_sPacket *p)
{
	p->cPacketType = 0x78;
	p->iLen = 3;
	packet_fill_seq(p);
	p->cID = _SID_Sidecar;
	packet_alloc_payload(p);
	p->cpPayload[0] = _PT_ChannelStatusReq;
	packet_fill_crc(p);
}

void packet_build_initialize_1(_sPacket *p)
{
	p->cPacketType = 0x70;
	p->iLen = 2;
	p->cSeq = 0x03;		/* Fixed seq. for init */
	p->cID = _SID_Sidecar;
	p->cpPayload = NULL;
	packet_fill_crc(p);
}

void packet_build_initialize_2(_sPacket *p)
{
	p->cPacketType = 0x78;
	p->iLen = 3;
	p->cSeq = 0x03;		/* Fixed seq. for init */
	p->cID = _SID_Sidecar;
	packet_alloc_payload(p);
	p->cpPayload[0] = 0x00;
	packet_fill_crc(p);
}

void packet_build_ack(_sPacket *in, _sPacket *out)
{
	if(in->cSeq & 0x80)
		warnx("building a response to an ACK!\n");
	out->cPacketType = in->cPacketType;
	out->iLen = 2;
	packet_fill_seq_ack(out);
	out->cID = _SID_Sidecar;
	out->cpPayload = NULL;
	packet_fill_crc(out);
}

void packet_build_keypress(_sPacket *p, uint8_t key)
{
	p->cPacketType = 0x78;
	p->iLen = 4;
	packet_fill_seq(p);
	p->cID = _SID_Sidecar;
	packet_alloc_payload(p);
	p->cpPayload[0] = _PT_IRKeypress;
	p->cpPayload[1] = key;
	packet_fill_crc(p);
}

void packet_fill_seq_ack(_sPacket *p)
{
	seq_ack = (seq_ack + 1) & 0x3;
	p->cSeq = 0x80 | (seq_req << 4) | seq_ack;
}

void packet_fill_seq(_sPacket *p)
{
	seq_req = (seq_req + 1) & 0x3;
	p->cSeq = (seq_req << 4) | seq_ack;
}

void packet_received(_sPacket *p)
{
	debug("RECV: type 0x%02x, len %d, seq 0x%02x\n",
	      p->cPacketType, p->iLen, p->cSeq);
	if(p->iLen>2) {
		int i;
		debug("      data");
		for(i=0;i<p->iLen-2;i++) 
			debug(" 0x%02x",p->cpPayload[i]);
		debug("\n");
	}		

	seq_ack = p->cSeq & 0x3;
	seq_req = (p->cSeq >> 4) & 0x3;
	if(!(p->cSeq & 0x80)) {
		_sPacket r;
		debug("Sending automatic ACK\n");
		packet_build_ack(p,&r);
		serial_sendpacket(&r);
		packet_cleanup(&r);
	}
}

void packet_sent(_sPacket *p)
{
	debug("SENT: type 0x%02x, len %d, seq 0x%02x\n",
	      p->cPacketType, p->iLen, p->cSeq);
	if(p->iLen>2) {
		int i;
		debug("      data");
		for(i=0;i<p->iLen-2;i++) 
			debug(" 0x%02x",p->cpPayload[i]);
		debug("\n");
	}		

	seq_ack = p->cSeq & 0x3;
	seq_req = (p->cSeq >> 4) & 0x3;
}

int packet_is_ack(_sPacket *p, _sPacket *ack)
{
	if((ack->iLen == 0x02) && 
	   (ack->cSeq == (((p->cSeq + 0x01) & 0x33) | 0x80)) &&
	   (ack->cPacketType == p->cPacketType))
		return 1;
	else
		return 0;
}

int packet_extract_channel(_sPacket *p)
{
	uint16_t chan;
	if(p->cPacketType!=0x78 || p->iLen<0x04 ||
	   p->cpPayload==NULL)
		return -1;
	if(p->cpPayload[0]!=_PT_ChannelStatus &&
	   p->cpPayload[0]!=_PT_ChannelStatusEx) {
		debug("wrong payload: 0x%02x\n",p->cpPayload[0]);
		return -1;
	}
	chan = (p->cpPayload[1] << 8) |
		p->cpPayload[2];
	return chan;
}
